@@ -14,47 +14,52 @@ use super::Uint128;
 ///
 /// The greatest possible value that can be represented is 340282366920938463463.374607431768211455 (which is (2^128 - 1) / 10^18)
 #[derive(Copy, Clone, Default, Debug, PartialEq, Eq, PartialOrd, Ord, JsonSchema)]
-pub struct Decimal(#[schemars(with = "String")] u128);
-
-const DECIMAL_FRACTIONAL: u128 = 1_000_000_000_000_000_000; // 1*10**18
-const DECIMAL_FRACTIONAL_SQUARED: u128 = 1_000_000_000_000_000_000_000_000_000_000_000_000; // (1*10**18)**2 = 1*10**36
+pub struct Decimal(#[schemars(with = "String")] Uint128);
 
 impl Decimal {
-    pub const MAX: Decimal = Decimal(u128::MAX);
+    const DECIMAL_FRACTIONAL: Uint128 = Uint128::new(1_000_000_000_000_000_000u128); // 1*10**18
+    const DECIMAL_FRACTIONAL_SQUARED: Uint128 =
+        Uint128::new(1_000_000_000_000_000_000_000_000_000_000_000_000u128); // (1*10**18)**2 = 1*10**36
+    const DECIMAL_PLACES: usize = 18; // This needs to be an even number.
+
+    pub const MAX: Self = Self(Uint128::MAX);
 
     /// Create a 1.0 Decimal
-    pub const fn one() -> Decimal {
-        Decimal(DECIMAL_FRACTIONAL)
+    pub const fn one() -> Self {
+        Decimal(Self::DECIMAL_FRACTIONAL)
     }
 
     /// Create a 0.0 Decimal
-    pub const fn zero() -> Decimal {
-        Decimal(0)
+    pub const fn zero() -> Self {
+        Decimal(Uint128::zero())
     }
 
     /// Convert x% into Decimal
-    pub fn percent(x: u64) -> Decimal {
-        Decimal((x as u128) * 10_000_000_000_000_000)
+    pub fn percent(x: u64) -> Self {
+        Decimal(((x as u128) * 10_000_000_000_000_000).into())
     }
 
     /// Convert permille (x/1000) into Decimal
-    pub fn permille(x: u64) -> Decimal {
-        Decimal((x as u128) * 1_000_000_000_000_000)
+    pub fn permille(x: u64) -> Self {
+        Decimal(((x as u128) * 1_000_000_000_000_000).into())
     }
 
     /// Returns the ratio (numerator / denominator) as a Decimal
-    pub fn from_ratio<A: Into<u128>, B: Into<u128>>(numerator: A, denominator: B) -> Decimal {
-        let numerator: u128 = numerator.into();
-        let denominator: u128 = denominator.into();
-        if denominator == 0 {
+    pub fn from_ratio(numerator: impl Into<Uint128>, denominator: impl Into<Uint128>) -> Self {
+        let numerator: Uint128 = numerator.into();
+        let denominator: Uint128 = denominator.into();
+        if denominator.is_zero() {
             panic!("Denominator must not be zero");
         }
-        // TODO: better algorithm with less rounding potential?
-        Decimal(numerator * DECIMAL_FRACTIONAL / denominator)
+
+        Decimal(
+            // numerator * DECIMAL_FRACTIONAL / denominator
+            numerator.multiply_ratio(Self::DECIMAL_FRACTIONAL, denominator),
+        )
     }
 
     pub fn is_zero(&self) -> bool {
-        self.0 == 0
+        self.0.is_zero()
     }
 
     /// Returns the approximate square root as a Decimal.
@@ -64,7 +69,11 @@ impl Decimal {
         // Algorithm described in https://hackmd.io/@webmaster128/SJThlukj_
         // We start with the highest precision possible and lower it until
         // there's no overflow.
-        (0..=9)
+        //
+        // TODO: This could be made more efficient once log10 is in:
+        // https://github.com/rust-lang/rust/issues/70887
+        // The max precision is something like `9 - log10(self.0) / 2`.
+        (0..=Self::DECIMAL_PLACES / 2)
             .rev()
             .find_map(|i| self.sqrt_with_precision(i))
             // The last step (i = 0) is guaranteed to succeed because `isqrt(u128::MAX) * 10^9` does not overflow
@@ -75,11 +84,13 @@ impl Decimal {
     /// Precision *must* be a number between 0 and 9 (inclusive).
     ///
     /// Returns `None` if the internal multiplication overflows.
-    fn sqrt_with_precision(&self, precision: u32) -> Option<Self> {
+    fn sqrt_with_precision(&self, precision: usize) -> Option<Self> {
+        let precision = precision as u32;
+
         let inner_mul = 100u128.pow(precision);
-        self.0.checked_mul(inner_mul).map(|inner| {
-            let outer_mul = 10u128.pow(9 - precision);
-            Decimal(inner.isqrt() * outer_mul)
+        self.0.checked_mul(inner_mul.into()).ok().map(|inner| {
+            let outer_mul = 10u128.pow(Self::DECIMAL_PLACES as u32 / 2 - precision);
+            Decimal(inner.isqrt().checked_mul(Uint128::from(outer_mul)).unwrap())
         })
     }
 }
@@ -87,25 +98,25 @@ impl Decimal {
 impl Fraction<u128> for Decimal {
     #[inline]
     fn numerator(&self) -> u128 {
-        self.0
+        self.0.u128()
     }
 
     #[inline]
     fn denominator(&self) -> u128 {
-        DECIMAL_FRACTIONAL
+        Self::DECIMAL_FRACTIONAL.u128()
     }
 
     /// Returns the multiplicative inverse `1/d` for decimal `d`.
     ///
     /// If `d` is zero, none is returned.
-    fn inv(&self) -> Option<Decimal> {
+    fn inv(&self) -> Option<Self> {
         if self.is_zero() {
             None
         } else {
             // Let self be p/q with p = self.0 and q = DECIMAL_FRACTIONAL.
             // Now we calculate the inverse a/b = q/p such that b = DECIMAL_FRACTIONAL. Then
             // `a = DECIMAL_FRACTIONAL*DECIMAL_FRACTIONAL / self.0`.
-            Some(Decimal(DECIMAL_FRACTIONAL_SQUARED / self.0))
+            Some(Decimal(Self::DECIMAL_FRACTIONAL_SQUARED / self.0))
         }
     }
 }
@@ -118,30 +129,38 @@ impl FromStr for Decimal {
     /// Disallowed: "", ".23"
     ///
     /// This never performs any kind of rounding.
-    /// More than 18 fractional digits, even zeros, result in an error.
+    /// More than DECIMAL_PLACES fractional digits, even zeros, result in an error.
     fn from_str(input: &str) -> Result<Self, Self::Err> {
         let mut parts_iter = input.split('.');
 
         let whole_part = parts_iter.next().unwrap(); // split always returns at least one element
         let whole = whole_part
-            .parse::<u128>()
+            .parse::<Uint128>()
             .map_err(|_| StdError::generic_err("Error parsing whole"))?;
         let mut atomics = whole
-            .checked_mul(DECIMAL_FRACTIONAL)
-            .ok_or_else(|| StdError::generic_err("Value too big"))?;
+            .checked_mul(Self::DECIMAL_FRACTIONAL)
+            .map_err(|_| StdError::generic_err("Value too big"))?;
 
         if let Some(fractional_part) = parts_iter.next() {
             let fractional = fractional_part
-                .parse::<u128>()
+                .parse::<Uint128>()
                 .map_err(|_| StdError::generic_err("Error parsing fractional"))?;
-            let exp = (18usize.checked_sub(fractional_part.len())).ok_or_else(|| {
-                StdError::generic_err("Cannot parse more than 18 fractional digits")
-            })?;
-            debug_assert!(exp <= 18);
-            let fractional_factor = 10u128.pow(exp as u32);
+            let exp =
+                (Self::DECIMAL_PLACES.checked_sub(fractional_part.len())).ok_or_else(|| {
+                    StdError::generic_err(format!(
+                        "Cannot parse more than {} fractional digits",
+                        Self::DECIMAL_PLACES
+                    ))
+                })?;
+            debug_assert!(exp <= Self::DECIMAL_PLACES);
+            let fractional_factor = Uint128::from(10u128.pow(exp as u32));
             atomics = atomics
-                .checked_add(fractional * fractional_factor)
-                .ok_or_else(|| StdError::generic_err("Value too big"))?;
+                .checked_add(
+                    // The inner multiplication can't overflow because
+                    // fractional < 10^DECIMAL_PLACES && fractional_factor <= 10^DECIMAL_PLACES
+                    fractional.checked_mul(fractional_factor).unwrap(),
+                )
+                .map_err(|_| StdError::generic_err("Value too big"))?;
         }
 
         if parts_iter.next().is_some() {
@@ -154,13 +173,14 @@ impl FromStr for Decimal {
 
 impl fmt::Display for Decimal {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let whole = (self.0) / DECIMAL_FRACTIONAL;
-        let fractional = (self.0) % DECIMAL_FRACTIONAL;
+        let whole = (self.0) / Self::DECIMAL_FRACTIONAL;
+        let fractional = (self.0).checked_rem(Self::DECIMAL_FRACTIONAL).unwrap();
 
-        if fractional == 0 {
+        if fractional.is_zero() {
             write!(f, "{}", whole)
         } else {
-            let fractional_string = format!("{:018}", fractional);
+            let fractional_string =
+                format!("{:0>padding$}", fractional, padding = Self::DECIMAL_PLACES);
             f.write_str(&whole.to_string())?;
             f.write_char('.')?;
             f.write_str(fractional_string.trim_end_matches('0'))?;
@@ -197,7 +217,7 @@ impl ops::Mul<Decimal> for Uint128 {
         if self.is_zero() || rhs.is_zero() {
             return Uint128::zero();
         }
-        self.multiply_ratio(rhs.0, DECIMAL_FRACTIONAL)
+        self.multiply_ratio(rhs.0, Decimal::DECIMAL_FRACTIONAL)
     }
 }
 
@@ -213,13 +233,13 @@ impl ops::Div<Uint128> for Decimal {
     type Output = Self;
 
     fn div(self, rhs: Uint128) -> Self::Output {
-        Decimal(self.0 / rhs.u128())
+        Decimal(self.0 / rhs)
     }
 }
 
 impl ops::DivAssign<Uint128> for Decimal {
     fn div_assign(&mut self, rhs: Uint128) {
-        self.0 /= rhs.u128();
+        self.0 /= rhs;
     }
 }
 
@@ -235,7 +255,7 @@ impl Serialize for Decimal {
 
 /// Deserializes as a base64 string
 impl<'de> Deserialize<'de> for Decimal {
-    fn deserialize<D>(deserializer: D) -> Result<Decimal, D::Error>
+    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
     where
         D: Deserializer<'de>,
     {
@@ -272,25 +292,25 @@ mod tests {
     #[test]
     fn decimal_one() {
         let value = Decimal::one();
-        assert_eq!(value.0, DECIMAL_FRACTIONAL);
+        assert_eq!(value.0, Decimal::DECIMAL_FRACTIONAL);
     }
 
     #[test]
     fn decimal_zero() {
         let value = Decimal::zero();
-        assert_eq!(value.0, 0);
+        assert!(value.0.is_zero());
     }
 
     #[test]
     fn decimal_percent() {
         let value = Decimal::percent(50);
-        assert_eq!(value.0, DECIMAL_FRACTIONAL / 2);
+        assert_eq!(value.0, Decimal::DECIMAL_FRACTIONAL / Uint128::from(2u8));
     }
 
     #[test]
     fn decimal_permille() {
         let value = Decimal::permille(125);
-        assert_eq!(value.0, DECIMAL_FRACTIONAL / 8);
+        assert_eq!(value.0, Decimal::DECIMAL_FRACTIONAL / Uint128::from(8u8));
     }
 
     #[test]
@@ -312,13 +332,22 @@ mod tests {
         // 1/3 (result floored)
         assert_eq!(
             Decimal::from_ratio(1u64, 3u64),
-            Decimal(333_333_333_333_333_333)
+            Decimal(Uint128::from(333_333_333_333_333_333u128))
         );
 
         // 2/3 (result floored)
         assert_eq!(
             Decimal::from_ratio(2u64, 3u64),
-            Decimal(666_666_666_666_666_666)
+            Decimal(Uint128::from(666_666_666_666_666_666u128))
+        );
+
+        // large inputs
+        assert_eq!(Decimal::from_ratio(0u128, u128::MAX), Decimal::zero());
+        assert_eq!(Decimal::from_ratio(u128::MAX, u128::MAX), Decimal::one());
+        // 340282366920938463463 is the largest integer <= Decimal::MAX
+        assert_eq!(
+            Decimal::from_ratio(340282366920938463463u128, 1u128),
+            Decimal::from_str("340282366920938463463").unwrap()
         );
     }
 
@@ -331,8 +360,8 @@ mod tests {
     #[test]
     fn decimal_implements_fraction() {
         let fraction = Decimal::from_str("1234.567").unwrap();
-        assert_eq!(fraction.numerator(), 1_234_567_000_000_000_000_000);
-        assert_eq!(fraction.denominator(), 1_000_000_000_000_000_000);
+        assert_eq!(fraction.numerator(), 1_234_567_000_000_000_000_000u128);
+        assert_eq!(fraction.denominator(), 1_000_000_000_000_000_000u128);
     }
 
     #[test]
@@ -358,14 +387,14 @@ mod tests {
         assert_eq!(Decimal::from_str("00.40").unwrap(), Decimal::percent(40));
         assert_eq!(Decimal::from_str("00.04").unwrap(), Decimal::percent(4));
 
-        // Can handle 18 fractional digits
+        // Can handle DECIMAL_PLACES fractional digits
         assert_eq!(
             Decimal::from_str("7.123456789012345678").unwrap(),
-            Decimal(7123456789012345678)
+            Decimal(Uint128::from(7123456789012345678u128))
         );
         assert_eq!(
             Decimal::from_str("7.999999999999999999").unwrap(),
-            Decimal(7999999999999999999)
+            Decimal(Uint128::from(7999999999999999999u128))
         );
 
         // Works for documented max value
@@ -420,7 +449,7 @@ mod tests {
     fn decimal_from_str_errors_for_more_than_18_fractional_digits() {
         match Decimal::from_str("7.1234567890123456789").unwrap_err() {
             StdError::GenericErr { msg, .. } => {
-                assert_eq!(msg, "Cannot parse more than 18 fractional digits")
+                assert_eq!(msg, "Cannot parse more than 18 fractional digits",)
             }
             e => panic!("Unexpected error: {:?}", e),
         }
@@ -535,7 +564,10 @@ mod tests {
     #[test]
     fn decimal_add() {
         let value = Decimal::one() + Decimal::percent(50); // 1.5
-        assert_eq!(value.0, DECIMAL_FRACTIONAL * 3 / 2);
+        assert_eq!(
+            value.0,
+            Decimal::DECIMAL_FRACTIONAL * Uint128::from(3u8) / Uint128::from(2u8)
+        );
     }
 
     #[test]
@@ -547,7 +579,7 @@ mod tests {
     #[test]
     fn decimal_sub() {
         let value = Decimal::one() - Decimal::percent(50); // 0.5
-        assert_eq!(value.0, DECIMAL_FRACTIONAL / 2);
+        assert_eq!(value.0, Decimal::DECIMAL_FRACTIONAL / Uint128::from(2u8));
     }
 
     #[test]
@@ -682,26 +714,77 @@ mod tests {
         // Decimals
         assert_eq!(Decimal::percent(125).to_string(), "1.25");
         assert_eq!(Decimal::percent(42638).to_string(), "426.38");
-        assert_eq!(Decimal::percent(1).to_string(), "0.01");
+        assert_eq!(Decimal::percent(3).to_string(), "0.03");
         assert_eq!(Decimal::permille(987).to_string(), "0.987");
 
-        assert_eq!(Decimal(1).to_string(), "0.000000000000000001");
-        assert_eq!(Decimal(10).to_string(), "0.00000000000000001");
-        assert_eq!(Decimal(100).to_string(), "0.0000000000000001");
-        assert_eq!(Decimal(1000).to_string(), "0.000000000000001");
-        assert_eq!(Decimal(10000).to_string(), "0.00000000000001");
-        assert_eq!(Decimal(100000).to_string(), "0.0000000000001");
-        assert_eq!(Decimal(1000000).to_string(), "0.000000000001");
-        assert_eq!(Decimal(10000000).to_string(), "0.00000000001");
-        assert_eq!(Decimal(100000000).to_string(), "0.0000000001");
-        assert_eq!(Decimal(1000000000).to_string(), "0.000000001");
-        assert_eq!(Decimal(10000000000).to_string(), "0.00000001");
-        assert_eq!(Decimal(100000000000).to_string(), "0.0000001");
-        assert_eq!(Decimal(10000000000000).to_string(), "0.00001");
-        assert_eq!(Decimal(100000000000000).to_string(), "0.0001");
-        assert_eq!(Decimal(1000000000000000).to_string(), "0.001");
-        assert_eq!(Decimal(10000000000000000).to_string(), "0.01");
-        assert_eq!(Decimal(100000000000000000).to_string(), "0.1");
+        assert_eq!(
+            Decimal(Uint128::from(1u128)).to_string(),
+            "0.000000000000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(10u128)).to_string(),
+            "0.00000000000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(100u128)).to_string(),
+            "0.0000000000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(1000u128)).to_string(),
+            "0.000000000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(10000u128)).to_string(),
+            "0.00000000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(100000u128)).to_string(),
+            "0.0000000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(1000000u128)).to_string(),
+            "0.000000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(10000000u128)).to_string(),
+            "0.00000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(100000000u128)).to_string(),
+            "0.0000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(1000000000u128)).to_string(),
+            "0.000000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(10000000000u128)).to_string(),
+            "0.00000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(100000000000u128)).to_string(),
+            "0.0000001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(10000000000000u128)).to_string(),
+            "0.00001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(100000000000000u128)).to_string(),
+            "0.0001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(1000000000000000u128)).to_string(),
+            "0.001"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(10000000000000000u128)).to_string(),
+            "0.01"
+        );
+        assert_eq!(
+            Decimal(Uint128::from(100000000000000000u128)).to_string(),
+            "0.1"
+        );
     }
 
     #[test]
