@@ -4,6 +4,7 @@ use std::convert::{TryFrom, TryInto};
 use std::fmt::{self};
 use std::iter::Sum;
 use std::ops;
+use std::str::FromStr;
 
 use crate::errors::{DivideByZeroError, OverflowError, OverflowOperation, StdError};
 use crate::{ConversionOverflowError, Uint256, Uint64};
@@ -184,7 +185,15 @@ impl TryFrom<&str> for Uint128 {
     type Error = StdError;
 
     fn try_from(val: &str) -> Result<Self, Self::Error> {
-        match val.parse::<u128>() {
+        Self::from_str(val)
+    }
+}
+
+impl FromStr for Uint128 {
+    type Err = StdError;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        match s.parse::<u128>() {
             Ok(u) => Ok(Uint128(u)),
             Err(e) => Err(StdError::generic_err(format!("Parsing u128: {}", e))),
         }
@@ -205,7 +214,7 @@ impl From<Uint128> for u128 {
 
 impl fmt::Display for Uint128 {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "{}", self.0)
+        self.0.fmt(f)
     }
 }
 
@@ -213,7 +222,11 @@ impl ops::Add<Uint128> for Uint128 {
     type Output = Self;
 
     fn add(self, rhs: Self) -> Self {
-        Uint128(self.u128().checked_add(rhs.u128()).unwrap())
+        Uint128(
+            self.u128()
+                .checked_add(rhs.u128())
+                .expect("attempt to add with overflow"),
+        )
     }
 }
 
@@ -221,7 +234,7 @@ impl<'a> ops::Add<&'a Uint128> for Uint128 {
     type Output = Self;
 
     fn add(self, rhs: &'a Uint128) -> Self {
-        Uint128(self.u128().checked_add(rhs.u128()).unwrap())
+        self + *rhs
     }
 }
 
@@ -229,7 +242,11 @@ impl ops::Sub<Uint128> for Uint128 {
     type Output = Self;
 
     fn sub(self, rhs: Self) -> Self {
-        Uint128(self.u128().checked_sub(rhs.u128()).unwrap())
+        Uint128(
+            self.u128()
+                .checked_sub(rhs.u128())
+                .expect("attempt to subtract with overflow"),
+        )
     }
 }
 
@@ -237,7 +254,27 @@ impl<'a> ops::Sub<&'a Uint128> for Uint128 {
     type Output = Self;
 
     fn sub(self, rhs: &'a Uint128) -> Self {
-        Uint128(self.u128().checked_sub(rhs.u128()).unwrap())
+        self - *rhs
+    }
+}
+
+impl ops::Mul<Uint128> for Uint128 {
+    type Output = Self;
+
+    fn mul(self, rhs: Self) -> Self::Output {
+        Self(
+            self.u128()
+                .checked_mul(rhs.u128())
+                .expect("attempt to multiply with overflow"),
+        )
+    }
+}
+
+impl<'a> ops::Mul<&'a Uint128> for Uint128 {
+    type Output = Self;
+
+    fn mul(self, rhs: &'a Uint128) -> Self::Output {
+        self.mul(*rhs)
     }
 }
 
@@ -245,7 +282,11 @@ impl ops::Div<Uint128> for Uint128 {
     type Output = Self;
 
     fn div(self, rhs: Self) -> Self::Output {
-        Self(self.u128().checked_div(rhs.u128()).unwrap())
+        Self(
+            self.u128()
+                .checked_div(rhs.u128())
+                .expect("attempt to divide by zero"),
+        )
     }
 }
 
@@ -253,7 +294,7 @@ impl<'a> ops::Div<&'a Uint128> for Uint128 {
     type Output = Self;
 
     fn div(self, rhs: &'a Uint128) -> Self::Output {
-        Self(self.u128().checked_div(rhs.u128()).unwrap())
+        self / *rhs
     }
 }
 
@@ -261,7 +302,11 @@ impl ops::Shr<u32> for Uint128 {
     type Output = Self;
 
     fn shr(self, rhs: u32) -> Self::Output {
-        Self(self.u128().checked_shr(rhs).unwrap())
+        Self(
+            self.u128()
+                .checked_shr(rhs)
+                .expect("attempt to shift right with overflow"),
+        )
     }
 }
 
@@ -269,55 +314,67 @@ impl<'a> ops::Shr<&'a u32> for Uint128 {
     type Output = Self;
 
     fn shr(self, rhs: &'a u32) -> Self::Output {
-        Self(self.u128().checked_shr(*rhs).unwrap())
+        self >> *rhs
     }
 }
 
 impl ops::AddAssign<Uint128> for Uint128 {
     fn add_assign(&mut self, rhs: Uint128) {
-        self.0 = self.0.checked_add(rhs.u128()).unwrap();
+        *self = *self + rhs;
     }
 }
 
 impl<'a> ops::AddAssign<&'a Uint128> for Uint128 {
     fn add_assign(&mut self, rhs: &'a Uint128) {
-        self.0 = self.0.checked_add(rhs.u128()).unwrap();
+        *self = *self + rhs;
     }
 }
 
 impl ops::SubAssign<Uint128> for Uint128 {
     fn sub_assign(&mut self, rhs: Uint128) {
-        self.0 = self.0.checked_sub(rhs.u128()).unwrap();
+        *self = *self - rhs;
     }
 }
 
 impl<'a> ops::SubAssign<&'a Uint128> for Uint128 {
     fn sub_assign(&mut self, rhs: &'a Uint128) {
-        self.0 = self.0.checked_sub(rhs.u128()).unwrap();
+        *self = *self - rhs;
+    }
+}
+
+impl ops::MulAssign<Uint128> for Uint128 {
+    fn mul_assign(&mut self, rhs: Self) {
+        *self = *self * rhs;
+    }
+}
+
+impl<'a> ops::MulAssign<&'a Uint128> for Uint128 {
+    fn mul_assign(&mut self, rhs: &'a Uint128) {
+        *self = *self * rhs;
     }
 }
 
 impl ops::DivAssign<Uint128> for Uint128 {
     fn div_assign(&mut self, rhs: Self) {
-        self.0 = self.0.checked_div(rhs.u128()).unwrap();
+        *self = *self / rhs;
     }
 }
 
 impl<'a> ops::DivAssign<&'a Uint128> for Uint128 {
     fn div_assign(&mut self, rhs: &'a Uint128) {
-        self.0 = self.0.checked_div(rhs.u128()).unwrap();
+        *self = *self / rhs;
     }
 }
 
 impl ops::ShrAssign<u32> for Uint128 {
     fn shr_assign(&mut self, rhs: u32) {
-        self.0 = self.0.checked_shr(rhs).unwrap();
+        *self = *self >> rhs;
     }
 }
 
 impl<'a> ops::ShrAssign<&'a u32> for Uint128 {
     fn shr_assign(&mut self, rhs: &'a u32) {
-        self.0 = self.0.checked_shr(*rhs).unwrap();
+        *self = *self >> rhs;
     }
 }
 
@@ -460,6 +517,12 @@ mod tests {
         assert_eq!(a.to_string(), "0");
     }
 
+    #[test]
+    fn uint128_display_padding_works() {
+        let a = Uint128::from(123u64);
+        assert_eq!(format!("Embedded: {:05}", a), "Embedded: 00123");
+    }
+
     #[test]
     fn uint128_is_zero_works() {
         assert!(Uint128::zero().is_zero());
@@ -545,11 +608,10 @@ mod tests {
         let base = Uint128(500);
 
         // factor 1/1
-        assert_eq!(base.multiply_ratio(1u128, 1u128), Uint128(500));
-        assert_eq!(base.multiply_ratio(3u128, 3u128), Uint128(500));
-        assert_eq!(base.multiply_ratio(654321u128, 654321u128), Uint128(500));
-        // Reactivate after https://github.com/CosmWasm/cosmwasm/issues/920
-        // assert_eq!(base.multiply_ratio(u128::MAX, u128::MAX), Uint128(500));
+        assert_eq!(base.multiply_ratio(1u128, 1u128), base);
+        assert_eq!(base.multiply_ratio(3u128, 3u128), base);
+        assert_eq!(base.multiply_ratio(654321u128, 654321u128), base);
+        assert_eq!(base.multiply_ratio(u128::MAX, u128::MAX), base);
 
         // factor 3/2
         assert_eq!(base.multiply_ratio(3u128, 2u128), Uint128(750));
