@@ -4,7 +4,7 @@ use std::fmt;
 
 use crate::{Binary, ContractResult};
 
-use super::{Attribute, CosmosMsg, Empty};
+use super::{CosmosMsg, Empty, Event};
 
 /// Use this to define when the contract gets a response callback.
 /// If you only need it for errors or success you can select just those in order
@@ -49,7 +49,7 @@ where
     T: Clone + fmt::Debug + PartialEq + JsonSchema,
 {
     /// new creates a "fire and forget" message with the pre-0.14 semantics
-    pub fn new<M: Into<CosmosMsg<T>>>(msg: M) -> Self {
+    pub fn new(msg: impl Into<CosmosMsg<T>>) -> Self {
         SubMsg {
             id: UNUSED_MSG_ID,
             msg: msg.into(),
@@ -59,17 +59,17 @@ where
     }
 
     /// create a `SubMsg` that will provide a `reply` with the given id if the message returns `Ok`
-    pub fn reply_on_success<M: Into<CosmosMsg<T>>>(msg: M, id: u64) -> Self {
+    pub fn reply_on_success(msg: impl Into<CosmosMsg<T>>, id: u64) -> Self {
         Self::reply_on(msg.into(), id, ReplyOn::Success)
     }
 
     /// create a `SubMsg` that will provide a `reply` with the given id if the message returns `Err`
-    pub fn reply_on_error<M: Into<CosmosMsg<T>>>(msg: M, id: u64) -> Self {
+    pub fn reply_on_error(msg: impl Into<CosmosMsg<T>>, id: u64) -> Self {
         Self::reply_on(msg.into(), id, ReplyOn::Error)
     }
 
     /// create a `SubMsg` that will always provide a `reply` with the given id
-    pub fn reply_always<M: Into<CosmosMsg<T>>>(msg: M, id: u64) -> Self {
+    pub fn reply_always(msg: impl Into<CosmosMsg<T>>, id: u64) -> Self {
         Self::reply_on(msg.into(), id, ReplyOn::Always)
     }
 
@@ -116,54 +116,3 @@ pub struct SubMsgExecutionResponse {
     pub events: Vec<Event>,
     pub data: Option<Binary>,
 }
-
-/// A full Cosmos SDK event as documented in
-/// https://docs.cosmos.network/v0.42/core/events.html.
-///
-/// This version uses string attributes (similar to
-/// https://github.com/cosmos/cosmos-sdk/blob/v0.42.5/proto/cosmos/base/abci/v1beta1/abci.proto#L56-L70),
-/// which then get magically converted to bytes for Tendermint somewhere between
-/// the Rust-Go interface, JSON deserialization and the `NewEvent` call in Cosmos SDK.
-#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, JsonSchema)]
-pub struct Event {
-    /// The event type. This is renamed to "ty" because "type" is reserved in Rust. This sucks, we know.
-    #[serde(rename = "type")]
-    pub ty: String,
-    pub attributes: Vec<Attribute>,
-}
-
-impl Event {
-    /// Create a new event with the given type and an empty list of attributes.
-    pub fn new(ty: impl Into<String>) -> Self {
-        Event {
-            ty: ty.into(),
-            attributes: Vec::with_capacity(10),
-        }
-    }
-
-    /// Add an attribute to the event.
-    pub fn attr(mut self, key: impl Into<String>, value: impl Into<String>) -> Self {
-        self.attributes.push(Attribute {
-            key: key.into(),
-            value: value.into(),
-        });
-        self
-    }
-}
-
-#[cfg(test)]
-mod tests {
-    use super::*;
-    use crate::attr;
-
-    #[test]
-    fn event_construction() {
-        let event_direct = Event {
-            ty: "test".to_string(),
-            attributes: vec![attr("foo", "bar"), attr("bar", "baz")],
-        };
-        let event_builder = Event::new("test").attr("foo", "bar").attr("bar", "baz");
-
-        assert_eq!(event_direct, event_builder);
-    }
-}
